@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.5 2001/03/28 18:24:25 hpa Exp $ */
+/* tftp-hpa: $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa Exp $ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -41,7 +41,7 @@ static const char *copyright =
  All rights reserved.\n";
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
-static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.5 2001/03/28 18:24:25 hpa Exp $";
+static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.7 2001/03/30 01:30:50 hpa Exp $";
 #endif /* not lint */
 
 /*
@@ -72,9 +72,18 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.5 2001/03/28 18:24:25 hpa
 #include <unistd.h>
 #include <limits.h>
 
+#include "../config.h"
 #include "tftpsubs.h"
 #include "recvfrom.h"
 
+void bsd_signal(int, void (*)(int));
+
+#ifndef HAVE_SIGSETJMP
+#define sigsetjmp(x,y)  setjmp(x)
+#define siglongjmp(x,y) longjmp(x,y)
+#define sigjmp_buf jmp_buf
+#endif
+
 #define	TIMEOUT 5		/* Default timeout (seconds) */
 #define TRIES   4		/* Number of attempts to send each packet */
 #define TIMEOUT_LIMIT (TRIES*(TRIES+1)/2)
@@ -86,7 +95,6 @@ static const char *rcsid = "tftp-hpa $Id: tftpd.c,v 1.5 2001/03/28 18:24:25 hpa
 #define EOPTNEG	8
 #endif
 
-extern	int errno;
 extern	char *__progname;
 int	peer;
 int	timeout    = TIMEOUT;
@@ -322,7 +330,7 @@ struct formats {
 } formats[] = {
 	{ "netascii",	validate_access,	sendfile,	recvfile, 1 },
 	{ "octet",	validate_access,	sendfile,	recvfile, 0 },
-	{ NULL }
+	{ NULL, NULL, NULL, NULL, 0 }
 };
 
 /*
@@ -625,7 +633,7 @@ validate_access(char *filename, int mode, struct formats *pf)
 }
 
 int	timeout;
-jmp_buf	timeoutbuf;
+sigjmp_buf	timeoutbuf;
 
 void
 timer(int sig)
@@ -634,7 +642,7 @@ timer(int sig)
 	timeout += rexmtval;
 	if (timeout >= maxtimeout)
 		exit(0);
-	longjmp(timeoutbuf, 1);
+	siglongjmp(timeoutbuf, 1);
 }
 
 /*
@@ -647,18 +655,18 @@ sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
 	struct tftphdr *ap;    /* ack packet */
 	int block = 1, size, n;
 
-	signal(SIGALRM, timer);
 	ap = (struct tftphdr *)ackbuf;
 
         if (oap) {
 	     timeout = 0;
-	     (void)setjmp(timeoutbuf);
+	     (void)sigsetjmp(timeoutbuf,1);
 oack:
 	     if (send(peer, oap, oacklen, 0) != oacklen) {
 		  syslog(LOG_ERR, "tftpd: oack: %m\n");
 		  goto abort;
 	     }
 	     for ( ; ; ) {
+	          bsd_signal(SIGALRM, timer);
 		  alarm(rexmtval);
 		  n = recv(peer, ackbuf, sizeof(ackbuf), 0);
 		  alarm(0);
@@ -694,7 +702,7 @@ oack:
 		dp->th_opcode = htons((u_short)DATA);
 		dp->th_block = htons((u_short)block);
 		timeout = 0;
-		(void) setjmp(timeoutbuf);
+		(void) sigsetjmp(timeoutbuf,1);
 
 send_data:
 		if (send(peer, dp, size + 4, 0) != size + 4) {
@@ -703,6 +711,7 @@ send_data:
 		}
 		read_ahead(file, pf->f_convert);
 		for ( ; ; ) {
+		        bsd_signal(SIGALRM, timer);
 			alarm(rexmtval);	/* read the ack */
 			n = recv(peer, ackbuf, sizeof (ackbuf), 0);
 			alarm(0);
@@ -751,12 +760,11 @@ recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
 	struct tftphdr *ap;    /* ack buffer */
 	int block = 0, acksize, n, size;
 
-	signal(SIGALRM, timer);
 	dp = w_init();
 	do {
 		timeout = 0;
 		
-		if (!block++ && oap) {
+		if (!block && oap) {
 		     ap = (struct tftphdr *)ackbuf;
 		     acksize = oacklen;
 		} else {
@@ -765,7 +773,9 @@ recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
 		     ap->th_block = htons((u_short)block);
 		     acksize = 4;
 		}
-		(void) setjmp(timeoutbuf);
+		block++;
+		(void) sigsetjmp(timeoutbuf,1);
+		bsd_signal(SIGALRM, timer);
 send_ack:
 		if (send(peer, ackbuf, acksize, 0) != acksize) {
 			syslog(LOG_ERR, "tftpd: write(ack): %m");
@@ -773,6 +783,7 @@ send_ack:
 		}
 		write_behind(file, pf->f_convert);
 		for ( ; ; ) {
+		        bsd_signal(SIGALRM, timer);
 			alarm(rexmtval);
 			n = recv(peer, dp, PKTSIZE, 0);
 			alarm(0);
@@ -809,7 +820,7 @@ send_ack:
 	ap->th_block = htons((u_short)(block));
 	(void) send(peer, ackbuf, 4, 0);
 
-	signal(SIGALRM, justquit);      /* just quit on timeout */
+	bsd_signal(SIGALRM, justquit);      /* just quit on timeout */
 	alarm(rexmtval);
 	n = recv(peer, buf, sizeof (buf), 0); /* normally times out and quits */
 	alarm(0);
